//! This module holds everything related to textures.

use std::cell::RefCell;
use std::rc::Rc;

use wasm_bindgen::prelude::*;
use wasm_bindgen::JsCast;

use web_sys::HtmlImageElement;

use crate::engine::bbox::Bbox;
use crate::{log, Result};

/// The size of sprites.
pub const SPRITE_SIZE: f64 = 32.0;

/// A wrapper to make it easier to manage HTML images.
pub struct Image(Rc<RefCell<InnerImage>>);

/// The content of the image.
pub struct InnerImage {
    /// The inner HtmlImageElement.
    pub inner: HtmlImageElement,

    /// Whether the image has been loaded or not.
    pub loaded: bool,
}

impl Image {
    /// Loads a new image from its src.
    pub fn new(path: &str) -> Result<Image> {
        let image = HtmlImageElement::new()?;
        image.set_src(path);

        let image = Image(Rc::new(RefCell::new(InnerImage {
            inner: image,
            loaded: false,
        })));

        let clone = image.0.clone();
        let path_clone = path.to_string();

        let cb = Closure::<dyn FnMut(_)>::new(move |_event: web_sys::Event| {
            log!("{} loaded", path_clone);
            (*clone).borrow_mut().loaded = true;
        });

        image
            .0
            .borrow()
            .inner
            .set_onload(Some(cb.as_ref().unchecked_ref()));

        cb.forget();

        Ok(image)
    }

    /// Returns whether the image is loaded.
    pub fn is_loaded(&self) -> bool {
        self.0.borrow().loaded
    }

    /// Renders the image on a context.
    pub fn render(
        &self,
        source: Bbox,
        dest: Bbox,
        context: &web_sys::CanvasRenderingContext2d,
    ) -> Result<()> {
        context.draw_image_with_html_image_element_and_sw_and_sh_and_dx_and_dy_and_dw_and_dh(
            &self.0.borrow().inner,
            source.position.x,
            source.position.y,
            source.size.x,
            source.size.y,
            dest.position.x,
            dest.position.y,
            dest.size.x,
            dest.size.y,
        )?;

        Ok(())
    }
}

/// The list of all our textures.
#[derive(Copy, Clone)]
pub enum Texture {
    /// The main character sprite.
    Rusty,

    /// The sprites from the overworld.
    Overworld,
}

/// Our texture manager.
///
/// It holds all our resources.
pub struct TextureManager {
    /// The main character sprite.
    pub rusty: Image,

    /// The sprites from the overworld.
    pub overworld: Image,
}

impl TextureManager {
    /// Creates and start the loading of all our textures.
    pub fn new() -> Result<TextureManager> {
        Ok(TextureManager {
            rusty: Image::new("static/textures/rusty.png")?,
            overworld: Image::new("static/textures/overworld.png")?,
        })
    }

    /// Returns the image corresponding to a certain texture.
    pub fn get(&self, texture: Texture) -> &Image {
        match texture {
            Texture::Rusty => &self.rusty,
            Texture::Overworld => &self.overworld,
        }
    }
}
